#pragma once

#include <fstream>
#include <map>
#include <vector>
#include <string>
#include "GasVertex.h"
#include "GasMaterial.h"
#include "GasBatch.h"
#include "GasNode.h"
#include "GasBone.h"
#include "AnimationHeader.h"

using namespace std;

class Exporter : public SceneExport
{
public:
	Exporter();
	~Exporter();

	/** Exports the scene. */
	int 			DoExport( const TCHAR* name, ExpInterface *ei, Interface* i, 
						BOOL suppressprompts=FALSE, DWORD options=0 );

	/** Show DLL's "About..." box */
	void			ShowAbout( HWND hwnd );

	/** Number of extensions supported */
	int				ExtCount();
	
	/** Extension #n (i.e. "3DS") */
	const TCHAR*	Ext( int n );					
	
	/** Long ASCII description (i.e. "Autodesk 3D Studio File") */
	const TCHAR*	LongDesc();
	
	/** Short ASCII description (i.e. "3D Studio") */
	const TCHAR*	ShortDesc();
	
	/** ASCII Author name */
	const TCHAR*	AuthorName();
	
	/** ASCII Copyright message */
	const TCHAR*	CopyrightMessage();
	
	/** Other message #1 */
	const TCHAR*	OtherMessage1();
	
	/** Other message #2 */
	const TCHAR*	OtherMessage2();
	
	/** Version number * 100 (i.e. v3.01 = 301) */
	unsigned int	Version();
	
	/** Returns TRUE if option supported. */
	BOOL			SupportsOptions( int ext, DWORD options );

	/** Processes the given node to collect mesh, material, and animation data
		Recursively collects data for all the node's children as well
	*/
	int ProcessNode(IGameNode* node, int index);

	/**
		Processes a single mesh to collect material, vertex, and normal data
	*/
	void ProcessMesh(IGameNode* node, GasNode* gasNode);

	/**
		Record animation data at the given time
		Stores data in gasNodes[ID]
	*/
	void GetTreeAnimAt(int ID, int currTime);

	/**
		Creates list of all children without processing them
		(Can collect each node's data again in a second pass
		 to avoid recomputing tranform data needlessly)
	*/
	int BuildNodeList(IGameNode* node);

	/**
		Get the length of the animation
	*/
	void GetAnimationTime(IGameScene* scene);

private:
	IGameScene*		m_igame;
	ofstream		file;

	void			deinit();

	Exporter( const Exporter& );
	Exporter& operator=( const Exporter& );

	int currTime;

	AnimationHeader animHeader;
	std::map<std::string, GasMaterial> materialMap;	//map of all unique materials in the scene
	std::vector<GasBatchStr> batches;				
	std::vector<GasNode*> gasnodes;					
	std::vector<IGameNode*> gamenodes;				

	/**
		Get the ibo index for the given vert in the batch vertex list
		Returns the index if found, -1 if not in the list
	*/
	int GetListIndex(GasVertex& vert, GasBatchStr& batch);
	int GetListIndex(GasSkinVertex& vert, GasBatchStr& batch);

	/**
		Gets the name of the texture without the filepath
		Texture name limited to 64 characters to save space
	*/
	void GetTextureName(std::string filepath, char* arr);

	/**
		Get the index of the specified material in the material list
		Returns either the index or -1 if not found
	*/
	int GetMaterialID(std::string materialName);

	/**
		Gets the node's transformation matrix (without any parent transformations)
	*/
	GMatrix GetMatrix(IGameNode* node, TimeValue time);

	void Transpose(GMatrix& mat);

	/**
		Get the node index in the node list
		Returns the index if found, -1 if not in the list
	*/
	int GetGlobalNodeIndex(IGameNode* node);
	int GetLocalNodeListIndex(IGameNode* node, std::vector<GasBone>& bones);

	/**
		Finds the 4 bones with the greatest weights and stores them in weights and bones 
		(last weight not recorded, but can be recomputed as 1 - (otherWeightTotal)
	*/
	void GetBonesAndWeights(std::vector< std::pair<float, int> >& weightList,
							Point3& weights, Point4& bones);

	/**
		Print out all collected information as a binary file

		Data is printed as
		VersionNumber
		Header:
			# of materials, offset to get to first material
			# of batches, offset to get to first batch
			# of nodes, offset to get to first node
			# of animations, offset to get to first animation
		All materials printed sequentially
		All batched printed sequentially
		All nodes printed sequentially
	*/
	void PrintHeader();			//describes layout of entire file
	void PrintMaterials();		//lists all materials
	void PrintVertices();		//lists all vertices
	void PrintBatches();		//lists all batches
	void PrintNodes();			//lists all nodes
	void PrintEOF();			//end the file

	/**
		Helper function to compute how large batches will be when written
		as binary data
	*/
	int ComputeBatchesSize();

	/** To reduce the final file size, only record trans/rot/scale info
		if there is an animation playing. No need to record matrices for
		times when the animation isn't playing
	*/
	bool ShouldExportTranslations(int nodeIndex);
	bool ShouldExportRotations(int nodeIndex);
	bool ShouldExportScales(int nodeIndex);
};